from collections import OrderedDict
from typing import Any, Dict, List, Optional, Set, Tuple

import numpy as np
import pandas as pd
import pyarrow as pa

from triad.collections.dict import IndexedOrderedDict
from triad.utils.assertion import assert_arg_not_none, assert_or_throw
from triad.utils.pandas_like import PD_UTILS
from triad.utils.pyarrow import (
    expression_to_schema,
    is_supported,
    schema_to_expression,
    to_pa_datatype,
    to_pandas_dtype,
)
from triad.utils.schema import (
    quote_name,
    safe_replace_out_of_quote,
    safe_search_out_of_quote,
    safe_split_and_unquote,
)


class SchemaError(Exception):
    """Exceptions related with construction and modifying schemas"""

    def __init__(self, message: Any):
        super().__init__(message)


class Schema(IndexedOrderedDict[str, pa.Field]):
    """A Schema wrapper on top of pyarrow.Fields. This has more features
    than pyarrow.Schema, and they can convert to each other.

    This class can be initialized from schema like objects. Here is a list of
    schema like objects:

    * pyarrow.Schema or Schema objects
    * pyarrow.Field: single field will be treated as a single column schema
    * schema expressions: :func:`~triad.utils.pyarrow.expression_to_schema`
    * Dict[str,Any]: key will be the columns, and value will be type like objects
    * Tuple[str,Any]: first item will be the only column name of the schema,
      and the second has to be a type like object
    * List[Any]: a list of Schema like objects
    * pandas.DataFrame: it will extract the dataframe's schema

    Here is a list of data type like objects:

    * pyarrow.DataType
    * pyarrow.Field: will only use the type attribute of the field
    * type expression or other objects: for :func:`~triad.utils.pyarrow.to_pa_datatype`

    .. admonition:: Examples

        .. code-block:: python

            Schema("a:int,b:int")
            Schema("a:int","b:int")
            Schema(a=int,b=str) # == Schema("a:long,b:str")
            Schema(dict(a=int,b=str)) # == Schema("a:long,b:str")
            Schema([(a,int),(b,str)]) # == Schema("a:long,b:str")
            Schema((a,int),(b,str)) # == Schema("a:long,b:str")
            Schema("a:[int],b:{x:int,y:{z:[str],w:byte}},c:[{x:str}]")

    .. note::

        * For supported pyarrow.DataTypes see :func:`~triad.utils.pyarrow.is_supported`
        * If you use python type as data type (e.g. `Schema(a=int,b=str)`) be aware
          the data type different. (e.g. python `int` type -> pyarrow `long`/`int64`
          type)
        * When not readonly, only `append` is allowed, `update` or `remove` are
          disallowed
        * When readonly, no modification on the existing schema is allowed
        * `append`, `update` and `remove` are always allowed when creating a new object
        * InvalidOperationError will be raised for disallowed operations
        * At most one of `*args` and `**kwargs` can be set

    :param args: one or multiple schema like objects, which will be combined in order
    :param kwargs: key value pairs for the schema
    """

    def __init__(self, *args: Any, **kwargs: Any):
        if len(args) > 0 and len(kwargs) > 0:
            raise SchemaError("Can't set both *args and **kwargs")
        if len(args) == 1:  # duplicate code for better performance
            if isinstance(args[0], Schema):
                super().__init__(args[0])  # type: ignore
                return
            fields: Optional[List[pa.Field]] = None
            if isinstance(args[0], str):
                fields = list(expression_to_schema(args[0]))
            if isinstance(args[0], pa.Schema):
                fields = list(args[0])
            if isinstance(args[0], pa.Field):
                fields = [args[0]]
            if fields is not None:
                fields = [self._validate_field(f) for f in fields]
                super().__init__([(x.name, x) for x in fields])
                return
        super().__init__()
        if len(args) > 0:
            self.append(list(args))
        elif len(kwargs) > 0:
            self.append(kwargs)

    @property
    def names(self) -> List[str]:
        """List of column names"""
        self._build_index()
        return self._index_key  # type: ignore

    @property
    def fields(self) -> List[pa.Field]:
        """List of pyarrow.Fields"""
        return list(self.values())

    @property
    def types(self) -> List[pa.DataType]:
        """List of pyarrow.DataTypes"""
        return [v.type for v in self.values()]

    @property
    def pyarrow_schema(self) -> pa.Schema:
        """convert as pyarrow.Schema"""
        return pa.schema(self.fields)

    @property
    def pa_schema(self) -> pa.Schema:
        """convert as pyarrow.Schema"""
        return self.pyarrow_schema

    @property
    def pandas_dtype(self) -> Dict[str, np.dtype]:
        """Convert as `dtype` dict for pandas dataframes.
        Currently, struct type is not supported
        """
        return self.to_pandas_dtype(self.pa_schema)

    def to_pandas_dtype(
        self, use_extension_types: bool = False, use_arrow_dtype: bool = False
    ) -> Dict[str, np.dtype]:
        """Convert as `dtype` dict for pandas dataframes.

        :param use_extension_types: if True, use pandas extension types,
            default False
        :param use_arrow_dtype: if True and when pandas supports ``ArrowDType``,
            use pyarrow types, default False

        .. note::

            * If ``use_extension_types`` is False and ``use_arrow_dtype`` is True,
                it converts all types to ``ArrowDType``
            * If both are true, it converts types to the numpy backend nullable
                dtypes if possible, otherwise, it converts to ``ArrowDType``
        """
        return to_pandas_dtype(
            self.pa_schema,
            use_extension_types=use_extension_types,
            use_arrow_dtype=use_arrow_dtype,
        )

    @property
    def pd_dtype(self) -> Dict[str, np.dtype]:
        """convert as `dtype` dict for pandas dataframes.
        Currently, struct type is not supported
        """
        return self.pandas_dtype

    def assert_not_empty(self) -> "Schema":
        """Raise exception if schema is empty"""
        if len(self) > 0:
            return self
        raise SchemaError("Schema can't be empty")

    def copy(self) -> "Schema":
        """Clone Schema object

        :return: cloned object
        """
        other = super().copy()
        assert isinstance(other, Schema)
        return other

    def __repr__(self) -> str:
        """Convert to the string representation of the schema"""
        return schema_to_expression(self.pyarrow_schema)

    def __iadd__(self, obj: Any) -> "Schema":
        """Append a schema like object to the current schema

        :param obj: a schema like object

        .. admonition:: Examples

            .. code-block:: python

                s = Schema("a:int,b:str")
                s += "c:long"
                s += ("d",int)
        """
        return self.append(obj)

    def __isub__(self, obj: Any) -> "Schema":
        """Remove columns from a schema is not allowed"""
        raise SchemaError("'-=' is not allowed for Schema")

    def __add__(self, obj: Any) -> "Schema":
        """Add a schema like object to the current schema

        :param obj: a schema like object

        .. admonition:: Examples

            .. code-block:: python

                s = Schema("a:int,b:str")
                s = s + "c:long" +  ("d",int)
        """
        return self.copy().append(obj)

    def __sub__(self, obj: Any) -> "Schema":
        """Remove columns or schema from the schema.

        :param obj: one column name, a list/set of column names or
            a schema like object
        :return: a schema excluding the columns in ``obj``

        .. note::

            If ``obj`` violates any of the following conditions, ``SchemaError``
            will be raised:

            * all columns in ``obj`` must be found in the schema
            * If ``obj`` is a schema like object, the types must also match
        """
        return self.remove(
            obj,
            ignore_key_mismatch=False,
            require_type_match=True,
            ignore_type_mismatch=False,
        )

    def __setitem__(  # type: ignore
        self, name: str, value: Any, *args: List[Any], **kwds: Dict[str, Any]
    ) -> None:
        assert_arg_not_none(value, "value")
        if name in self:  # update existing value is not allowed
            raise SchemaError(f"{name} already exists in {self}")
        if isinstance(value, pa.Field):
            assert_or_throw(
                name == value.name, SchemaError(f"{name} doesn't match {value}")
            )
        elif isinstance(value, pa.DataType):
            value = pa.field(name, value)
        else:
            value = pa.field(name, to_pa_datatype(value))
        assert_or_throw(
            is_supported(value.type), SchemaError(f"{value} is not supported")
        )
        super().__setitem__(name, value, *args, **kwds)  # type: ignore

    def __eq__(self, other: Any) -> bool:
        """Check if the two schemas are equal

        :param other: a schema like object
        :return: True if the two schemas are equal
        """
        if other is None:
            return False
        if other is self:
            return True
        if isinstance(other, Schema):
            return super().__eq__(other)
        if isinstance(other, str):
            return self.__repr__() == other
        return self == Schema(other)

    def __contains__(self, key: Any) -> bool:  # noqa: C901
        """Check if the schema contains the key.

        :param key: a column name, a list of column names, a data
            type like object or a schema like object
        :return: True if the schema contains the object
        """
        if key is None:
            return False
        if isinstance(key, str):
            if ":" not in key:
                return super().__contains__(key)
        elif isinstance(key, pa.Field):
            res = super().get(key.name, None)
            if res is None:
                return False
            return key.type == res.type
        elif isinstance(key, Schema):
            for f in key.values():
                if f not in self:
                    return False
            return True
        elif isinstance(key, List):
            for f in key:
                if f not in self:
                    return False
            return True
        return Schema(key) in self

    def append(self, obj: Any) -> "Schema":  # noqa: C901
        """Append schema like object to the current schema. Only new columns
        are allowed.

        :raises SchemaError: if a column exists or is invalid or obj is not convertible
        :return: the Schema object itself
        """
        try:
            if obj is None:
                return self
            elif isinstance(obj, pa.Field):
                self[obj.name] = obj.type
            elif isinstance(obj, str):
                self._append_pa_schema(expression_to_schema(obj))
            elif isinstance(obj, Dict):
                for k, v in obj.items():
                    self[k] = v
            elif isinstance(obj, pa.Schema):
                self._append_pa_schema(obj)
            elif isinstance(obj, pd.DataFrame):
                self._append_pa_schema(PD_UTILS.to_schema(obj))
            elif isinstance(obj, Tuple):  # type: ignore
                self[obj[0]] = obj[1]
            elif isinstance(obj, List):
                for x in obj:
                    self.append(x)
            else:
                raise SchemaError(f"Invalid schema to add {obj}")
            return self
        except SchemaError:
            raise
        except Exception as e:
            raise SchemaError(str(e))

    def remove(  # noqa: C901
        self,
        obj: Any,
        ignore_key_mismatch: bool = False,
        require_type_match: bool = True,
        ignore_type_mismatch: bool = False,
    ) -> "Schema":
        """Remove columns or schema from the schema

        :param obj: one column name, a list/set of column names or
            a schema like object
        :param ignore_key_mismatch: if True, ignore the non-existing keys,
            default False
        :param require_type_match: if True, a match requires the same key
            and same type (if ``obj`` contains type), otherwise, only the
            key needs to match, default True
        :param ignore_type_mismatch: if False, when keys match but types don't
            (if ``obj`` contains type), raise an exception ``SchemaError``,
            default False
        :return: a schema excluding the columns in ``obj``
        """
        if obj is None:
            return self.copy()
        target = self
        if isinstance(obj, str):
            if ":" in obj:  # expression
                ps = expression_to_schema(obj)
                pairs: List[Tuple[str, pa.DataType]] = list(zip(ps.names, ps.types))
            else:
                pairs = [(obj, None)]  # single key
        elif isinstance(obj, (pa.Schema, Schema)):
            pairs = list(zip(obj.names, obj.types))
        elif isinstance(obj, (List, Set)):
            keys: List[str] = []
            other: List[Any] = []
            for x in obj:
                if isinstance(x, str) and ":" not in x:
                    keys.append(x)
                else:
                    other.append(x)
            pairs = [(x, None) for x in keys]
            for o in other:
                target = target.remove(
                    o,
                    ignore_key_mismatch=ignore_key_mismatch,
                    require_type_match=require_type_match,
                    ignore_type_mismatch=ignore_type_mismatch,
                )
        else:
            return self.remove(
                Schema(obj),
                ignore_key_mismatch=ignore_key_mismatch,
                require_type_match=require_type_match,
                ignore_type_mismatch=ignore_type_mismatch,
            )
        od = OrderedDict(target)
        for k, v in pairs:
            # k = k.strip()
            # if k == "":
            #    continue
            if k not in od:
                if ignore_key_mismatch:
                    continue
                raise SchemaError(
                    f"Can't remove {quote_name(k)} from {target} (not found)"
                )
            if v is None:
                del od[k]
            else:
                tp = od[k].type
                if not require_type_match or tp == v:
                    del od[k]
                elif not ignore_type_mismatch:
                    raise SchemaError(
                        f"Unable to remove {k}:{v} from {self}, type mismatch"
                    )
        return Schema(od)

    def alter(self, subschema: Any) -> "Schema":
        """Alter the schema with a subschema

        :param subschema: a schema like object
        :return: the altered schema
        """
        if subschema is None:
            return self
        sub = Schema(subschema)
        assert_or_throw(
            sub.names in self,
            lambda: ValueError(f"{sub.names} are not all in {self}"),
        )
        return Schema([(k, sub.get(k, v)) for k, v in self.items()])

    def extract(  # noqa: C901
        self,
        obj: Any,
        ignore_key_mismatch: bool = False,
        require_type_match: bool = True,
        ignore_type_mismatch: bool = False,
    ) -> "Schema":
        """Extract a sub schema from the schema based on the columns in ``obj``

        :param obj: one column name, a list/set of column names or
            a schema like object
        :param ignore_key_mismatch: if True, ignore the non-existing keys,
            default False
        :param require_type_match: if True, a match requires the same key
            and same type (if ``obj`` contains type), otherwise, only the
            key needs to match, default True
        :param ignore_type_mismatch: if False, when keys match but types don't
            (if ``obj`` contains type), raise an exception ``SchemaError``,
            default False
        :return: a sub-schema containing the columns in ``obj``
        """
        if obj is None:
            return Schema()
        if isinstance(obj, str):
            if ":" in obj:  # expression
                ps = expression_to_schema(obj)
                pairs: List[Tuple[str, pa.DataType]] = list(zip(ps.names, ps.types))
            else:
                pairs = [(obj, None)]  # single key
        elif isinstance(obj, (pa.Schema, Schema)):
            pairs = list(zip(obj.names, obj.types))
        elif isinstance(obj, List):
            fields: List[pa.Field] = []
            for x in obj:
                if isinstance(x, str) and ":" not in x:
                    if x not in self:
                        if not ignore_key_mismatch:
                            raise SchemaError(f"Can't extract {x} from {self}")
                    else:
                        fields.append(self[x])
                else:
                    fields += self.extract(
                        x,
                        ignore_key_mismatch=ignore_key_mismatch,
                        require_type_match=require_type_match,
                        ignore_type_mismatch=ignore_type_mismatch,
                    ).fields
            return Schema(pa.schema(fields))
        else:
            return self.extract(
                Schema(obj),
                ignore_key_mismatch=ignore_key_mismatch,
                require_type_match=require_type_match,
                ignore_type_mismatch=ignore_type_mismatch,
            )
        fields = []
        for k, v in pairs:
            if k not in self:
                if ignore_key_mismatch:
                    continue
                raise SchemaError(f"Can't extract {k} from {self}")
            if v is None:
                fields.append(self[k])
            else:
                tp = self[k].type
                if not require_type_match or tp == v:
                    fields.append(self[k])
                elif not ignore_type_mismatch:
                    raise SchemaError(
                        f"Unable to extract {k}:{v} from {self}, type mismatch"
                    )
        return Schema(pa.schema(fields))

    def exclude(
        self,
        other: Any,
        require_type_match: bool = True,
        ignore_type_mismatch: bool = False,
    ) -> "Schema":
        """Exclude columns from the current schema which are also in ``other``.
        ``other`` can contain columns that are not in the current schema, they
        will be ignored.

        :param other: one column name, a list/set of column names or
            a schema like object
        :param require_type_match: if True, a match requires the same key
            and same type (if ``obj`` contains type), otherwise, only the
            key needs to match, default True
        :param ignore_type_mismatch: if False, when keys match but types don't
            (if ``obj`` contains type), raise an exception ``SchemaError``,
            default False
        :return: a schema excluding the columns in ``other``
        """
        return self.remove(
            other,
            ignore_key_mismatch=True,
            require_type_match=require_type_match,
            ignore_type_mismatch=ignore_type_mismatch,
        )

    def intersect(
        self,
        other: Any,
        require_type_match: bool = True,
        ignore_type_mismatch: bool = True,
        use_other_order: bool = False,
    ) -> "Schema":
        """Extract the sub-schema from the current schema which are also in
        ``other``. ``other`` can contain columns that are not in the current schema,
        they will be ignored.

        :param other: one column name, a list/set of column names or
            a schema like object
        :param require_type_match: if True, a match requires the same key
            and same type (if ``obj`` contains type), otherwise, only the
            key needs to match, default True
        :param ignore_type_mismatch: if False, when keys match but types don't
            (if ``obj`` contains type), raise an exception ``SchemaError``,
            default False
        :param use_other_order: if True, the output schema will use the column order
            of ``other``, default False
        :return: the intersected schema
        """
        if not use_other_order:
            diff = self.exclude(
                other,
                require_type_match=require_type_match,
                ignore_type_mismatch=ignore_type_mismatch,
            )
            return self - diff
        else:
            return self.extract(
                other,
                require_type_match=require_type_match,
                ignore_type_mismatch=ignore_type_mismatch,
            )

    def union(self, other: Any, require_type_match: bool = False) -> "Schema":
        """Union the ``other`` schema

        :param other: a schema like object
        :param require_type_match: if True, a match requires the same key
            and same type (if ``obj`` contains type), otherwise, only the
            key needs to match, default True
        :return: the new unioned schema
        """
        return self.copy().union_with(other, require_type_match=require_type_match)

    def union_with(self, other: Any, require_type_match: bool = False) -> "Schema":
        """Union the ``other`` schema into the current schema

        :param other: a schema like object
        :param require_type_match: if True, a match requires the same key
            and same type (if ``obj`` contains type), otherwise, only the
            key needs to match, default True
        :return: the current schema
        """
        if not isinstance(other, Schema):
            other = Schema(other)
        try:
            other = other.exclude(
                self, require_type_match=require_type_match, ignore_type_mismatch=False
            )
            self += other
            return self
        except Exception as e:
            raise SchemaError(f"Unable to union {self} with {other}: {str(e)}")

    def rename(self, columns: Dict[str, str], ignore_missing: bool = False) -> "Schema":
        """Rename the current schema and generate a new one

        :param columns: dictionary to map from old to new column names
        :return: renamed schema object
        """
        if not ignore_missing:
            for x in columns:
                if x not in self:
                    raise SchemaError(f"Failed to rename: {x} not in {self}")
        pairs = [
            (k if k not in columns else columns[k], v.type) for k, v in self.items()
        ]
        return Schema(pairs)

    def transform(self, *args: Any, **kwargs: Any) -> "Schema":
        """Transform the current schema to a new schema

        :raises SchemaError: if there is any exception
        :return: transformed schema

        .. admonition:: Examples

            .. code-block:: python

                s=Schema("a:int,b:int,c:str")
                s.transform("x:str") # x:str
                # add
                s.transform("*,x:str") # a:int,b:int,c:str,x:str
                s.transform("*","x:str") # a:int,b:int,c:str,x:str
                s.transform("*",x=str) # a:int,b:int,c:str,x:str
                # subtract
                s.transform("*-c,a") # b:int
                s.transform("*-c-a") # b:int
                s.transform("*~c,a,x") # b:int  # ~ means exlcude if exists
                s.transform("*~c~a~x") # b:int  # ~ means exlcude if exists
                # + means overwrite existing and append new
                s.transform("*+e:str,b:str,d:str") # a:int,b:str,c:str,e:str,d:str
                # you can have multiple operations
                s.transform("*+b:str-a") # b:str,c:str
                # callable
                s.transform(lambda s:s.fields[0]) # a:int
                s.transform(lambda s:s.fields[0], lambda s:s.fields[2]) # a:int,c:str
        """
        try:
            result = Schema()
            for a in args:
                if callable(a):
                    result += a(self)
                elif isinstance(a, str):
                    op_pos = [x[0] for x in safe_search_out_of_quote(a, "-~+")]
                    op_pos.append(len(a))
                    s = Schema(
                        safe_replace_out_of_quote(a[: op_pos[0]], "*", str(self))
                    )
                    for i in range(0, len(op_pos) - 1):
                        op, expr = a[op_pos[i]], a[(op_pos[i] + 1) : op_pos[i + 1]]
                        if op in ["-", "~"]:
                            cols = safe_split_and_unquote(
                                expr, ",", on_unquoted_empty="ignore"
                            )
                            s = s.exclude(cols) if op == "~" else s - cols
                        else:  # +
                            overwrite = Schema(expr)
                            s = Schema(
                                [(k, overwrite.get(k, v)) for k, v in s.items()]
                            ).union_with(overwrite)
                    result += s
                else:
                    result += a
            return result + Schema(kwargs)
        except SchemaError:
            raise
        except Exception as e:
            raise SchemaError(e)

    def create_empty_arrow_table(self) -> pa.Table:
        """Create an empty pyarrow table based on the schema"""
        if not hasattr(pa.Table, "from_pylist"):  # pragma: no cover
            arr = [pa.array([])] * len(self)
            return pa.Table.from_arrays(arr, schema=self.pa_schema)
        return pa.Table.from_pylist([], schema=self.pa_schema)

    def create_empty_pandas_df(
        self, use_extension_types: bool = False, use_arrow_dtype: bool = False
    ) -> pd.DataFrame:
        """Create an empty pandas dataframe based on the schema

        :param use_extension_types: if True, use pandas extension types,
            default False
        :param use_arrow_dtype: if True and when pandas supports ``ArrowDType``,
            use pyarrow types, default False
        :return: empty pandas dataframe
        """
        dtypes = self.to_pandas_dtype(
            use_extension_types=use_extension_types,
            use_arrow_dtype=use_arrow_dtype,
        )
        return pd.DataFrame({k: pd.Series(dtype=v) for k, v in dtypes.items()})

    def _pre_update(self, op: str, need_reindex: bool = True) -> None:
        if op == "__setitem__":
            super()._pre_update(op, need_reindex)
        else:
            raise SchemaError(f"{op} is not allowed in Schema")

    def _append_pa_schema(self, other: pa.Schema) -> "Schema":
        for k, v in zip(other.names, other.types):
            self[k] = v
        return self

    def _validate_field(self, field: pa.Field) -> pa.Field:
        assert_or_throw(
            is_supported(field.type), SchemaError(f"{field} type is not supported")
        )
        return field



import os
import sys
from io import BytesIO

import pytest

import triad.utils.io as iou


def test_join(tmpdir):
    assert iou.join(str(tmpdir)) == os.path.join(str(tmpdir))
    assert iou.join(str(tmpdir), "a", "b") == os.path.join(str(tmpdir), "a", "b")
    assert iou.join("dummy://", "a", "b", "c/") == "dummy://a/b/c"
    assert iou.join("dummy://a/", "b/", "c/") == "dummy://a/b/c"
    assert iou.join("dummy://a/", "b/", "*.parquet") == "dummy://a/b/*.parquet"


def test_glob(tmpdir):
    def _assert(gres, expected):
        for a, b in zip(gres, expected):
            assert iou.exists(a) and iou.exists(b)

    assert iou.glob(os.path.join(str(tmpdir), "a")) == []
    assert iou.glob(os.path.join(str(tmpdir), "a", "*.txt")) == []
    iou.touch(os.path.join(str(tmpdir), "a.txt"))
    _assert(
        iou.glob(os.path.join(str(tmpdir), "a.txt")),
        [os.path.join(str(tmpdir), "a.txt")],
    )
    _assert(
        iou.glob(os.path.join(str(tmpdir), "*.txt")),
        [os.path.join(str(tmpdir), "a.txt")],
    )
    iou.touch(os.path.join(str(tmpdir), "a", "a.txt"), auto_mkdir=True)
    _assert(
        iou.glob(os.path.join(str(tmpdir), "a", "*.txt")),
        [os.path.join(str(tmpdir), "a", "a.txt")],
    )

    iou.touch("memory://gtest/m.txt", auto_mkdir=True)
    assert iou.glob("memory://gtest/*.txt") == ["memory:///gtest/m.txt"]


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not a test for windows")
def test_join_not_win():
    assert iou.join("a", "b", "c/") == "a/b/c"
    assert iou.join("/a", "b", "c/") == "/a/b/c"
    assert iou.join("/a", "b", "*.parquet") == "/a/b/*.parquet"


@pytest.mark.skipif(
    not sys.platform.startswith("win"), reason="a test only for windows"
)
def test_join_is_win():
    assert iou.join("a", "b", "c") == "a\\b\\c"
    assert iou.join("c:\\a", "b", "c") == "c:\\a\\b\\c"
    assert iou.join("c:\\a", "b", "*.parquet") == "c:\\a\\b\\*.parquet"


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not a test for windows")
def test_abs_path_not_win(tmpdir):
    with iou.chdir(str(tmpdir)):
        assert iou.abs_path("a") == os.path.join(str(tmpdir), "a")
    assert iou.abs_path("/tmp/x") == "/tmp/x"
    assert iou.abs_path("file:///tmp/x") == "/tmp/x"
    assert iou.abs_path("memory://tmp/x") == "memory://tmp/x"
    assert iou.abs_path("memory:///tmp/x") == "memory:///tmp/x"
    assert iou.abs_path("dummy:///tmp/x") == "dummy:///tmp/x"


@pytest.mark.skipif(
    not sys.platform.startswith("win"), reason="a test only for windows"
)
def test_abs_path_is_win(tmpdir):
    with iou.chdir(str(tmpdir)):
        assert iou.abs_path("a") == os.path.join(str(tmpdir), "a")
    assert iou.abs_path("c:\\tmp\\x") == "c:\\tmp\\x"
    assert iou.abs_path("file://c:/tmp/x") == "c:\\tmp\\x"
    assert iou.abs_path("memory://tmp/x") == "memory://tmp/x"
    assert iou.abs_path("memory:///tmp/x") == "memory:///tmp/x"
    assert iou.abs_path("dummy:///tmp/x") == "dummy:///tmp/x"


def test_makedirs(tmpdir):
    path = os.path.join(str(tmpdir), "temp", "a")
    assert path == iou.makedirs(path, exist_ok=False)
    assert iou.exists(path)
    with pytest.raises(OSError):
        iou.makedirs(path, exist_ok=False)
    iou.makedirs(path, exist_ok=True)

    with iou.chdir(str(tmpdir)):
        assert os.path.join(str(tmpdir), "temp", "b") == iou.makedirs(
            iou.join("temp", "b"), exist_ok=False
        )
        assert iou.exists(iou.join("temp", "b"))
        assert iou.exists(os.path.join(str(tmpdir), "temp", "b"))

    path = "memory://temp/a"
    assert path == iou.makedirs(path, exist_ok=True)
    assert iou.exists(path)
    with pytest.raises(OSError):
        iou.makedirs(path, exist_ok=False)

    path = os.path.join(str(tmpdir), "temp", "aa")
    fs, _path = iou.url_to_fs(path)
    _path = fs.unstrip_protocol(_path)  # file:///...
    iou.makedirs(_path, exist_ok=False)
    assert iou.exists(path)


def test_exists(tmpdir):
    iou.write_text(os.path.join(str(tmpdir), "a.txt"), "a")
    assert iou.exists(os.path.join(str(tmpdir), "a.txt"))
    assert not iou.exists(os.path.join(str(tmpdir), "b.txt"))
    iou.write_text(os.path.join(str(tmpdir), "temp", "a.txt"), "a")
    assert iou.exists(os.path.join(str(tmpdir), "temp", "a.txt"))
    assert iou.exists(os.path.join(str(tmpdir), "temp"))
    assert iou.exists(str(tmpdir))

    with iou.chdir(str(tmpdir)):
        assert iou.exists("a.txt")
        assert iou.exists(os.path.join("temp", "a.txt"))
        assert iou.exists("temp")
        assert iou.exists(".")


def test_is_dir_or_file(tmpdir):
    for base in [str(tmpdir), "memory://"]:
        path = os.path.join(base, "a.txt")
        assert not iou.isfile(path)
        assert not iou.isdir(path)
        iou.write_text(path, "a")
        assert iou.isfile(path)
        assert not iou.isdir(path)
        path = os.path.join(base, "b")
        iou.makedirs(path)
        assert not iou.isfile(path)
        assert iou.isdir(path)


def test_touch(tmpdir):
    for base in [str(tmpdir), "memory://"]:
        path = os.path.join(base, "a.txt")
        iou.touch(path)
        assert iou.isfile(path)

        iou.write_text(path, "a")
        assert iou.read_text(path) == "a"
        iou.touch(path)
        assert iou.isfile(path)
        # assert iou.read_text(path) == ""

        if not base.startswith("memory"):
            pytest.raises(OSError, lambda: iou.touch(iou.join(base, "b", "c.txt")))

        iou.touch(iou.join(base, "b", "c.txt"), auto_mkdir=True)
        iou.exists(iou.join(base, "b", "c.txt"))


def test_rm(tmpdir):
    path = os.path.join(str(tmpdir), "a.txt")
    for recursive in [True, False]:
        iou.write_text(os.path.join(str(tmpdir), "a.txt"), "a")
        assert iou.exists(path)
        iou.rm(path, recursive=recursive)
        assert not iou.exists(path)

    path = os.path.join(str(tmpdir), "b", "c")
    iou.write_text(os.path.join(path, "a.txt"), "a")
    assert iou.exists(path)
    iou.rm(path, recursive=True)
    assert not iou.exists(path)


def test_read_write_content(tmpdir):
    iou.write_text(os.path.join(str(tmpdir), "temp", "a.txt"), "aa")
    assert iou.read_text(os.path.join(str(tmpdir), "temp", "a.txt")) == "aa"
    iou.write_bytes(os.path.join(str(tmpdir), "temp", "a.bin"), b"aa")
    assert iou.read_bytes(os.path.join(str(tmpdir), "temp", "a.bin")) == b"aa"


def test_zip_unzip(tmpdir):
    bio = BytesIO()
    with iou.zip_temp(bio) as tmpdir:
        iou.write_text(os.path.join(tmpdir, "a.txt"), "a")
        iou.write_text(os.path.join(tmpdir, "b.txt"), "b")

    with iou.unzip_to_temp(BytesIO(bio.getvalue())) as tmpdir:
        assert iou.read_text(os.path.join(tmpdir, "a.txt")) == "a"
        assert iou.read_text(os.path.join(tmpdir, "b.txt")) == "b"

    tf = os.path.join(str(tmpdir), "temp", "x.zip")
    with iou.zip_temp(tf) as tmpdir:
        iou.write_text(os.path.join(tmpdir, "a.txt"), "a")
        iou.write_text(os.path.join(tmpdir, "b.txt"), "b")

    with iou.unzip_to_temp(tf) as tmpdir:
        assert iou.read_text(os.path.join(tmpdir, "a.txt")) == "a"
        assert iou.read_text(os.path.join(tmpdir, "b.txt")) == "b"



import json
import pickle
from datetime import date, datetime

import numpy as np
import pandas as pd
import pyarrow as pa
import pytest
from pytest import raises

from triad.utils.pandas_like import _DEFAULT_DATETIME, PD_UTILS
from triad.utils.pyarrow import (
    PYARROW_VERSION,
    expression_to_schema,
    pa_table_to_pandas,
)

from .._utils import assert_df_eq


def test_to_schema():
    df = pd.DataFrame([[1.0, 2], [2.0, 3]])
    raises(ValueError, lambda: PD_UTILS.to_schema(df))
    df = pd.DataFrame([[1.0, 2], [2.0, 3]], columns=["x", "y"])
    assert list(pa.Schema.from_pandas(df)) == list(PD_UTILS.to_schema(df))
    df = pd.DataFrame([["a", 2], ["b", 3]], columns=["x", "y"])
    assert list(pa.Schema.from_pandas(df)) == list(PD_UTILS.to_schema(df))
    df = pd.DataFrame([], columns=["x", "y"])
    df = df.astype(dtype={"x": np.int32, "y": np.dtype("object")})
    assert [pa.field("x", pa.int32()), pa.field("y", pa.string())] == list(
        PD_UTILS.to_schema(df)
    )
    df = pd.DataFrame([[1, "x"], [2, "y"]], columns=["x", "y"])
    df = df.astype(dtype={"x": np.int32, "y": np.dtype("object")})
    assert list(pa.Schema.from_pandas(df)) == list(PD_UTILS.to_schema(df))
    df = pd.DataFrame([[1, "x"], [2, "y"]], columns=["x", "y"])
    df = df.astype(dtype={"x": np.int32, "y": np.dtype(str)})
    assert list(pa.Schema.from_pandas(df)) == list(PD_UTILS.to_schema(df))
    df = pd.DataFrame([[1, "x"], [2, "y"]], columns=["x", "y"])
    df = df.astype(dtype={"x": np.int32, "y": np.dtype("str")})
    assert list(pa.Schema.from_pandas(df)) == list(PD_UTILS.to_schema(df))

    df = pd.Series([], name="col", dtype="string").to_frame()
    assert expression_to_schema("col:str") == PD_UTILS.to_schema(df)

    df = pd.Series([], name="col", dtype="datetime64[ns]").to_frame()
    assert expression_to_schema("col:datetime") == PD_UTILS.to_schema(df)

    df = pd.Series([], name="col", dtype="datetime64[ns, UTC]").to_frame()
    assert expression_to_schema("col:timestamp(us, UTC)") == PD_UTILS.to_schema(df)

    # timestamp test
    df = pd.DataFrame(
        [[datetime(2020, 1, 1, 2, 3, 4, 5), date(2020, 2, 2)]], columns=["a", "b"]
    )
    assert list(expression_to_schema("a:datetime,b:date")) == list(
        PD_UTILS.to_schema(df)
    )

    # test index
    df = pd.DataFrame([[3.0, 2], [2.0, 3]], columns=["x", "y"])
    df = df.sort_values(["x"])
    assert list(pa.Schema.from_pandas(df, preserve_index=False)) == list(
        PD_UTILS.to_schema(df)
    )
    df.index.name = "x"
    raises(ValueError, lambda: PD_UTILS.to_schema(df))
    df = df.reset_index(drop=True)
    assert list(pa.Schema.from_pandas(df)) == list(PD_UTILS.to_schema(df))
    df["p"] = "p"
    df = df.set_index(["p"])
    df.index.name = None
    raises(ValueError, lambda: PD_UTILS.to_schema(df))


def test_as_array_iterable():
    df = DF([], "a:str,b:int")
    assert [] == df.as_array()
    assert [] == df.as_array(type_safe=True)

    df = DF([["a", 1]], "a:str,b:int")
    assert [["a", 1]] == df.as_array()
    assert [["a", 1]] == df.as_array(["a", "b"])
    assert [[1, "a"]] == df.as_array(["b", "a"])
    assert [[1, "a"]] == df.as_array(["b", "a"], null_schema=True)

    # prevent pandas auto type casting
    df = DF([[1.0, 1.1]], "a:double,b:int")
    assert [[1.0, 1]] == df.as_array()
    np.issubdtype(df.as_array()[0][0], np.float_)
    np.issubdtype(df.as_array()[0][1], np.integer)
    assert [[1.0, 1]] == df.as_array(["a", "b"])
    assert [[1, 1.0]] == df.as_array(["b", "a"])

    df = DF([[np.float64(1.0), 1.1]], "a:double,b:int")
    assert [[1.0, 1]] == df.as_array()
    np.issubdtype(df.as_array()[0][0], np.float_)
    np.issubdtype(df.as_array()[0][1], np.integer)

    df = DF([[pd.Timestamp("2020-01-01"), 1.1]], "a:datetime,b:int")
    assert [[datetime(2020, 1, 1), 1]] == df.as_array()
    assert isinstance(df.as_array(type_safe=True)[0][0], datetime)
    assert isinstance(df.as_array(type_safe=True)[0][1], int)

    df = DF([[pd.NaT, 1.1]], "a:datetime,b:int")
    assert df.as_array(type_safe=True)[0][0] is None
    assert isinstance(df.as_array(type_safe=True)[0][1], int)

    df = DF([[1.0, 1.1]], "a:double,b:int")
    assert [[1.0, 1]] == df.as_array(type_safe=True)
    np.issubdtype(df.as_array()[0][0], np.float_)
    np.issubdtype(df.as_array()[0][1], np.integer)


def test_as_array_iterable_datetime():
    df = pd.DataFrame(
        [[datetime(2020, 1, 1, 2, 3, 4, 5), date(2020, 2, 2)]], columns=["a", "b"]
    )
    v1 = list(PD_UTILS.as_array_iterable(df, type_safe=True))[0]
    v2 = list(
        PD_UTILS.as_array_iterable(
            df, schema=expression_to_schema("a:datetime,b:date"), type_safe=True
        )
    )[0]
    assert v1[0] == v2[0]
    assert not isinstance(v1[0], pd.Timestamp)
    assert type(v1[0]) == datetime
    assert type(v1[0]) == type(v2[0])

    assert v1[1] == v2[1]
    assert not isinstance(v1[1], pd.Timestamp)
    assert type(v1[1]) == date
    assert type(v1[1]) == type(v2[1])


@pytest.mark.skipif(PYARROW_VERSION.major < 8, reason="pyarrow>=8.0.0 required")
def test_nested():
    data = [[dict(b=[30, 40]), 2, 2.2]]
    df = DF(data, "a:{b:[int]},b:int,c:double")
    res = df.as_array(type_safe=True)
    assert res == data
    assert isinstance(res[0][1], int)
    assert not isinstance(res[0][1], np.number)
    assert isinstance(res[0][2], float)
    assert not isinstance(res[0][2], np.number)


def test_binary():
    b = pickle.dumps("xyz")
    data = [[b, b"xy"]]
    s = expression_to_schema("a:bytes,b:bytes")
    df = DF(data, "a:bytes,b:bytes")
    a = df.as_array(type_safe=True)
    assert [[b, b"xy"]] == a


def test_nan_none():
    df = DF([[None, None]], "b:str,c:double", True)
    assert pd.isna(df.native.iloc[0, 0])
    arr = df.as_array(type_safe=True)[0]
    assert arr[0] is None
    assert arr[1] is None

    df = DF([[None, None]], "b:int,c:bool", True)
    arr = df.as_array(type_safe=True)[0]
    assert arr[0] is None
    assert arr[1] is None

    df = DF([], "b:str,c:double", True)
    assert len(df.as_array()) == 0


def test_cast_df():
    df = pd.DataFrame(
        dict(
            a=pd.Series([1, 2], dtype="int32"), b=pd.Series([True, False], dtype="bool")
        )
    )
    assert df is PD_UTILS.cast_df(
        df,
        expression_to_schema("a:int32,b:bool"),
        use_extension_types=False,
        use_arrow_dtype=False,
    )
    res = PD_UTILS.cast_df(
        df,
        expression_to_schema("a:long,b:str"),
        use_extension_types=True,
        use_arrow_dtype=False,
    )
    assert isinstance(res.a.dtype, pd.Int64Dtype)
    assert isinstance(res.b.dtype, pd.StringDtype)
    assert res.values.tolist() == [[1, "true"], [2, "false"]]

    df = pd.DataFrame(dict(a=pd.Series(dtype="int32"), b=pd.Series(dtype="bool")))
    res = PD_UTILS.cast_df(
        df,
        expression_to_schema("a:long,b:str"),
        use_extension_types=True,
        use_arrow_dtype=False,
    )
    assert isinstance(res.a.dtype, pd.Int64Dtype)
    assert isinstance(res.b.dtype, pd.StringDtype)
    assert len(res) == 0


def test_boolean_enforce():
    df = DF([[1, True], [2, False], [3, None]], "b:int,c:bool", True)
    arr = df.as_array(type_safe=True)
    assert [[1, True], [2, False], [3, None]] == arr

    df = DF([[1, 1], [2, 0]], "b:int,c:bool", True)
    arr = df.as_array(type_safe=True)
    assert [[1, True], [2, False]] == arr

    df = DF([[1, "trUe"], [2, "False"], [3, None]], "b:int,c:bool", True)
    arr = df.as_array(type_safe=True)
    assert [[1, True], [2, False], [3, None]] == arr


def test_timestamp_enforce():
    df = DF(
        [
            [1, pd.Timestamp("2020-01-02 00:00:00", tz="UTC")],
            [2, pd.Timestamp("2020-01-03 00:00:00", tz="UTC")],
        ],
        "b:int,c:datetime",
        True,
    )
    arr = df.as_array(type_safe=True)
    assert [
        [1, pd.Timestamp("2020-01-02 00:00:00")],
        [2, pd.Timestamp("2020-01-03 00:00:00")],
    ] == arr

    df = DF(
        [
            [1, pd.Timestamp("2020-01-02 00:00:00")],
            [2, pd.Timestamp("2020-01-03 00:00:00")],
        ],
        "b:int,c:timestamp(ns, UTC)",
        True,
    )
    arr = df.as_array(type_safe=True)
    assert [
        [1, pd.Timestamp("2020-01-02 00:00:00", tz="UTC")],
        [2, pd.Timestamp("2020-01-03 00:00:00", tz="UTC")],
    ] == arr

    df = DF(
        [
            [1, pd.Timestamp("2020-01-02 00:00:00", tz="US/Pacific")],
            [2, pd.Timestamp("2020-01-03 00:00:00", tz="US/Pacific")],
        ],
        "b:int,c:timestamp(ns, UTC)",
        True,
    )
    arr = df.as_array(type_safe=True)
    assert [
        [1, pd.Timestamp("2020-01-02 08:00:00", tz="UTC")],
        [2, pd.Timestamp("2020-01-03 08:00:00", tz="UTC")],
    ] == arr

    df = DF(
        [
            [1, "2020-01-02 00:00:00"],
            [2, "2020-01-03 00:00:00"],
        ],
        "b:int,c:datetime",
        True,
    )
    arr = df.as_array(type_safe=True)
    assert [
        [1, pd.Timestamp("2020-01-02 00:00:00")],
        [2, pd.Timestamp("2020-01-03 00:00:00")],
    ] == arr

    df = DF(
        [
            [1, "2020-01-02 00:00:00-0500"],
            [2, "2020-01-03 00:00:00-0500"],
        ],
        "b:int,c:timestamp(ns, UTC)",
        True,
    )
    arr = df.as_array(type_safe=True)
    assert [
        [1, pd.Timestamp("2020-01-02 05:00:00", tz="UTC")],
        [2, pd.Timestamp("2020-01-03 05:00:00", tz="UTC")],
    ] == arr

    if hasattr(pd, "ArrowDtype"):
        df = pd.DataFrame(dict(a=pd.Series(["2020-01-02"], dtype="string[pyarrow]")))
        res = PD_UTILS.cast_df(df, expression_to_schema("a:datetime"))
        assert res.a.iloc[0] == pd.Timestamp("2020-01-02")


def test_fillna_default():
    df = pd.DataFrame([[1.0], [None]], columns=["x"])
    s = PD_UTILS.fillna_default(df["x"])
    assert [1.0, 0.0] == s.tolist()

    df = pd.DataFrame([["a"], [None]], columns=["x"])
    s = PD_UTILS.fillna_default(df["x"])
    assert ["a", ""] == s.tolist()

    df = pd.DataFrame([["a"], ["b"]], columns=["x"])
    s = PD_UTILS.fillna_default(df["x"].astype(np.str_))
    assert ["a", "b"] == s.tolist()

    dt = datetime.now()
    df = pd.DataFrame([[dt], [None]], columns=["x"])
    s = PD_UTILS.fillna_default(df["x"])
    assert [dt, _DEFAULT_DATETIME] == s.tolist()

    df = pd.DataFrame([[True], [None]], columns=["x"])
    s = PD_UTILS.fillna_default(df["x"])
    assert [True, ""] == s.tolist()

    df = pd.DataFrame([[True], [False]], columns=["x"])
    s = PD_UTILS.fillna_default(df["x"].astype(bool))
    assert [True, False] == s.tolist()


def test_safe_group_by_apply():
    df = DF([["a", 1], ["a", 2], [None, 3]], "b:str,c:long", True)

    def _m1(df):
        PD_UTILS.ensure_compatible(df)
        df["ct"] = df.shape[0]
        return df

    res = PD_UTILS.safe_groupby_apply(df.native, ["b"], _m1)
    PD_UTILS.ensure_compatible(res)
    assert 3 == res.shape[0]
    assert 3 == res.shape[1]
    assert [["a", 1, 2], ["a", 2, 2], [None, 3, 1]] == res.astype(object).where(
        ~pd.isna(res), None
    ).values.tolist()

    res = PD_UTILS.safe_groupby_apply(df.native, [], _m1)
    PD_UTILS.ensure_compatible(res)
    assert 3 == res.shape[0]
    assert 3 == res.shape[1]
    assert [["a", 1, 3], ["a", 2, 3], [None, 3, 3]] == res.astype(object).where(
        ~pd.isna(res), None
    ).values.tolist()

    df = DF([[1.0, "a"], [1.0, "b"], [None, "c"], [None, "d"]], "b:double,c:str", True)
    res = PD_UTILS.safe_groupby_apply(df.native, ["b"], _m1)
    assert [
        [1.0, "a", 2],
        [1.0, "b", 2],
        [pd.NA, "c", 2],
        [pd.NA, "d", 2],
    ].__repr__() == res.values.tolist().__repr__()


def test_to_parquet_friendly():
    if hasattr(pd, "ArrowDtype"):
        adf = pa.Table.from_pandas(pd.DataFrame(dict(a=["a", "b"], c=[1, 2])))
        pdf = pa_table_to_pandas(adf, use_extension_types=True, use_arrow_dtype=True)
        res = PD_UTILS.to_parquet_friendly(pdf)
        assert res is pdf

        adf = pa.Table.from_pandas(pd.DataFrame(dict(a=[["a", "b"], ["c"]], c=[1, 2])))
        pdf = pa_table_to_pandas(adf, use_extension_types=False, use_arrow_dtype=True)
        res = PD_UTILS.to_parquet_friendly(pdf)
        assert res.dtypes["a"] == np.dtype(object)
        assert res.dtypes["c"] == pd.ArrowDtype(pa.int64())
        pdf = pa_table_to_pandas(adf, use_extension_types=True, use_arrow_dtype=True)
        res = PD_UTILS.to_parquet_friendly(pdf)
        assert res.dtypes["a"] == np.dtype(object)
        assert res.dtypes["c"] == pd.Int64Dtype()
        res = PD_UTILS.to_parquet_friendly(pdf, partition_cols=["c"])
        assert res.dtypes["a"] == np.dtype(object)
        assert res.dtypes["c"] == np.dtype(object)


def test_safe_group_by_apply_special_types():
    def _m1(df):
        PD_UTILS.ensure_compatible(df)
        df["ct"] = df.shape[0]
        return df

    df = DF(
        [["a", 1.0], [None, 3.0], [None, 3.0], [None, None]], "a:str,b:double", True
    )
    res = PD_UTILS.safe_groupby_apply(df.native, ["a", "b"], _m1)
    PD_UTILS.ensure_compatible(res)
    assert 4 == res.shape[0]
    assert 3 == res.shape[1]
    DF(
        [["a", 1.0, 1], [None, 3.0, 2], [None, 3.0, 2], [None, None, 1]],
        "a:str,b:double,ct:int",
        True,
    ).assert_eq(res)

    dt = datetime.now()
    df = DF([["a", dt], [None, dt], [None, dt], [None, None]], "a:str,b:datetime", True)
    res = PD_UTILS.safe_groupby_apply(df.native, ["a", "b"], _m1)
    PD_UTILS.ensure_compatible(res)
    assert 4 == res.shape[0]
    assert 3 == res.shape[1]
    DF(
        [["a", dt, 1], [None, dt, 2], [None, dt, 2], [None, None, 1]],
        "a:str,b:datetime,ct:int",
        True,
    ).assert_eq(res)

    dt = date(2020, 1, 1)
    df = DF([["a", dt], [None, dt], [None, dt], [None, None]], "a:str,b:date", True)
    res = PD_UTILS.safe_groupby_apply(df.native, ["a", "b"], _m1)
    PD_UTILS.ensure_compatible(res)
    assert 4 == res.shape[0]
    assert 3 == res.shape[1]
    DF(
        [["a", dt, 1], [None, dt, 2], [None, dt, 2], [None, None, 1]],
        "a:str,b:date,ct:int",
        True,
    ).assert_eq(res)

    dt = date(2020, 1, 1)
    df = DF([["a", dt], ["b", dt], ["b", dt], ["b", None]], "a:str,b:date", True)
    res = PD_UTILS.safe_groupby_apply(df.native, ["a", "b"], _m1)
    PD_UTILS.ensure_compatible(res)
    assert 4 == res.shape[0]
    assert 3 == res.shape[1]
    DF(
        [["a", dt, 1], ["b", dt, 2], ["b", dt, 2], ["b", None, 1]],
        "a:str,b:date,ct:int",
        True,
    ).assert_eq(res)


def test_is_compatible_index():
    df = DF([["a", 1], [None, 2]], "a:str,b:int", True)
    assert PD_UTILS.is_compatile_index(df.native)
    tdf = df.native.sort_values("a")
    assert PD_UTILS.is_compatile_index(tdf)
    tdf = tdf.set_index("a")
    assert not PD_UTILS.is_compatile_index(tdf)


def test_parse_join_types():
    assert "cross" == PD_UTILS.parse_join_type("CROss")
    assert "inner" == PD_UTILS.parse_join_type("join")
    assert "inner" == PD_UTILS.parse_join_type("Inner")
    assert "left_outer" == PD_UTILS.parse_join_type("left")
    assert "left_outer" == PD_UTILS.parse_join_type("left  outer")
    assert "right_outer" == PD_UTILS.parse_join_type("right")
    assert "right_outer" == PD_UTILS.parse_join_type("right_ outer")
    assert "full_outer" == PD_UTILS.parse_join_type("full")
    assert "full_outer" == PD_UTILS.parse_join_type(" outer ")
    assert "full_outer" == PD_UTILS.parse_join_type("full_outer")
    assert "left_anti" == PD_UTILS.parse_join_type("anti")
    assert "left_anti" == PD_UTILS.parse_join_type("left anti")
    assert "left_semi" == PD_UTILS.parse_join_type("semi")
    assert "left_semi" == PD_UTILS.parse_join_type("left semi")
    raises(
        NotImplementedError,
        lambda: PD_UTILS.parse_join_type("right semi"),
    )


def test_drop_duplicates():
    def assert_eq(df, expected, expected_cols):
        res = PD_UTILS.drop_duplicates(df)
        assert_df_eq(res, expected, expected_cols)

    a = _to_df([["x", "a"], ["x", "a"], [None, None]], ["a", "b"])
    assert_eq(a, [["x", "a"], [None, None]], ["a", "b"])


def test_union():
    def assert_eq(df1, df2, unique, expected, expected_cols):
        res = PD_UTILS.union(df1, df2, unique=unique)
        assert_df_eq(res, expected, expected_cols)

    a = _to_df([["x", "a"], ["x", "a"], [None, None]], ["a", "b"])
    b = _to_df([["xx", "aa"], [None, None], ["a", "x"]], ["b", "a"])
    assert_eq(
        a,
        b,
        False,
        [
            ["x", "a"],
            ["x", "a"],
            [None, None],
            ["xx", "aa"],
            [None, None],
            ["a", "x"],
        ],
        ["a", "b"],
    )
    assert_eq(
        a,
        b,
        True,
        [["x", "a"], ["xx", "aa"], [None, None], ["a", "x"]],
        ["a", "b"],
    )


def test_intersect():
    def assert_eq(df1, df2, unique, expected, expected_cols):
        res = PD_UTILS.intersect(df1, df2, unique=unique)
        assert_df_eq(res, expected, expected_cols)

    a = _to_df([["x", "a"], ["x", "a"], [None, None]], ["a", "b"])
    b = _to_df([["xx", "aa"], [None, None], [None, None], ["a", "x"]], ["b", "a"])
    assert_eq(a, b, False, [[None, None]], ["a", "b"])
    assert_eq(a, b, True, [[None, None]], ["a", "b"])
    b = _to_df([["xx", "aa"], [None, None], ["x", "a"]], ["b", "a"])
    assert_eq(a, b, False, [["x", "a"], ["x", "a"], [None, None]], ["a", "b"])
    assert_eq(a, b, True, [["x", "a"], [None, None]], ["a", "b"])


def test_except():
    def assert_eq(df1, df2, unique, expected, expected_cols):
        res = PD_UTILS.except_df(df1, df2, unique=unique)
        assert_df_eq(res, expected, expected_cols)

    a = _to_df([["x", "a"], ["x", "a"], [None, None]], ["a", "b"])
    b = _to_df([["xx", "aa"], [None, None], ["a", "x"]], ["b", "a"])
    assert_eq(a, b, False, [["x", "a"], ["x", "a"]], ["a", "b"])
    assert_eq(a, b, True, [["x", "a"]], ["a", "b"])
    b = _to_df([["xx", "aa"], [None, None], ["x", "a"]], ["b", "a"])
    assert_eq(a, b, False, [], ["a", "b"])
    assert_eq(a, b, True, [], ["a", "b"])


def test_joins():
    def assert_eq(df1, df2, join_type, on, expected, expected_cols):
        res = PD_UTILS.join(df1, df2, join_type=join_type, on=on)
        assert_df_eq(res, expected, expected_cols)

    df1 = _to_df([[0, 1], [2, 3]], ["a", "b"])
    df2 = _to_df([[0, 10], [20, 30]], ["a", "c"])
    df3 = _to_df([[0, 1], [None, 3]], ["a", "b"])
    df4 = _to_df([[0, 10], [None, 30]], ["a", "c"])
    assert_eq(df1, df2, "inner", ["a"], [[0, 1, 10]], ["a", "b", "c"])
    assert_eq(df3, df4, "inner", ["a"], [[0, 1, 10]], ["a", "b", "c"])
    assert_eq(df1, df2, "left_semi", ["a"], [[0, 1]], ["a", "b"])
    assert_eq(df3, df4, "left_semi", ["a"], [[0, 1]], ["a", "b"])
    assert_eq(df1, df2, "left_anti", ["a"], [[2, 3]], ["a", "b"])
    assert_eq(df3, df4, "left_anti", ["a"], [[None, 3]], ["a", "b"])
    assert_eq(
        df1,
        df2,
        "left_outer",
        ["a"],
        [[0, 1, 10], [2, 3, None]],
        ["a", "b", "c"],
    )
    assert_eq(
        df3,
        df4,
        "left_outer",
        ["a"],
        [[0, 1, 10], [None, 3, None]],
        ["a", "b", "c"],
    )
    assert_eq(
        df1,
        df2,
        "right_outer",
        ["a"],
        [[0, 1, 10], [20, None, 30]],
        ["a", "b", "c"],
    )
    assert_eq(
        df3,
        df4,
        "right_outer",
        ["a"],
        [[0, 1, 10], [None, None, 30]],
        ["a", "b", "c"],
    )
    assert_eq(
        df1,
        df2,
        "full_outer",
        ["a"],
        [[0, 1, 10], [2, 3, None], [20, None, 30]],
        ["a", "b", "c"],
    )
    assert_eq(
        df3,
        df4,
        "full_outer",
        ["a"],
        [[0, 1, 10], [None, 3, None], [None, None, 30]],
        ["a", "b", "c"],
    )

    df1 = _to_df([[0, 1], [None, 3]], ["a", "b"])
    df2 = _to_df([[0, 10], [None, 30]], ["c", "d"])
    assert_eq(
        df1,
        df2,
        "cross",
        [],
        [
            [0, 1, 0, 10],
            [None, 3, 0, 10],
            [0, 1, None, 30],
            [None, 3, None, 30],
        ],
        ["a", "b", "c", "d"],
    )


def _to_df(data, cols):
    return pd.DataFrame(data, columns=cols)


class DF:  # This is a mock
    def __init__(self, data, schema, enforce=False):
        s = expression_to_schema(schema)
        df = pd.DataFrame(data, columns=s.names)
        self.native = PD_UTILS.cast_df(
            df, s, use_extension_types=True, use_arrow_dtype=False
        )
        self.schema = s

    def as_array(self, cols=None, type_safe=False, null_schema=False):
        schema = None if null_schema else self.schema
        return list(
            PD_UTILS.as_array_iterable(
                self.native, schema=schema, columns=cols, type_safe=type_safe
            )
        )

    def assert_eq(self, df):
        pd.testing.assert_frame_equal(self.native, df, check_dtype=False)
